@@ -4,8 +4,8 @@
 
 # Standard library imports
 import os
+import struct
 from io import BytesIO, BufferedIOBase
-from math import sqrt, exp
 from typing import Dict, List, Optional, BinaryIO, Any, Tuple, Union
 
 # Third party imports
@@ -13,6 +13,9 @@ import numpy as np
 from nptyping import Array
 from PIL import Image
 
+# Self imports
+from flyr.thermogram import FlyrThermogram
+
 
 # Constants
 segment_sep = b"\xff"
@@ -36,7 +39,21 @@ chunk_metadata_length = (
 )
 
 
-def unpack(path_or_stream: Union[str, BinaryIO]) -> Array[float, ..., ...]:
+def unpack(path_or_stream: Union[str, BinaryIO]) -> FlyrThermogram:
+    """ Unpacks the FLIR image, meaning that it will return the thermal data
+        embedded in the image.
+
+        Parameters
+        ----------
+        path_or_stream : Union[str, BinaryIO]
+            Either a path (string) to a FLIR file, or a byte stream such as
+            BytesIO or file opened as `open(file_path, "rb")`.
+
+        Returns
+        -------
+        FlyrThermogram
+            When successful, a FlyrThermogram object containing thermogram data.
+    """
     if isinstance(path_or_stream, str) and os.path.isfile(path_or_stream):
         with open(path_or_stream, "rb") as flirh:
             return unpack(flirh)
@@ -44,31 +61,32 @@ def unpack(path_or_stream: Union[str, BinaryIO]) -> Array[float, ..., ...]:
         stream = path_or_stream
         flir_app1_stream = extract_flir_app1(stream)
         flir_records = parse_flir_app1(flir_app1_stream)
-        thermal = parse_thermal(flir_app1_stream, flir_records)
-        return thermal
+        thermogram = parse_thermal(flir_app1_stream, flir_records)
+
+        return thermogram
     else:
         raise ValueError("Incorrect input")  # TODO improved error message
 
 
 def extract_flir_app1(stream: BinaryIO) -> BinaryIO:
     """ Extracts the FLIR APP1 bytes.
-    
+
         Parameters
         ---------
         stream : BinaryIO
             A full bytes stream of a JPEG file, expected to be a FLIR file.
-        
+
         Raises
         ------
         ValueError
-            When the file is invalid in one the next ways, a 
+            When the file is invalid in one the next ways, a
             ValueError is thrown.
-            
+
             * File is not a JPEG
             * A FLIR chunk number occurs more than once
             * The total chunks count is inconsistent over multiple chunks
             * No APP1 segments are successfully parsed
-        
+
         Returns
         -------
         BinaryIO
@@ -83,7 +101,10 @@ def extract_flir_app1(stream: BinaryIO) -> BinaryIO:
     chunks: Dict[int, bytes] = {}
     while True:
         b = stream.read(1)
-        if not b == segment_sep:
+        if b == b"":
+            break
+
+        if b != segment_sep:
             continue
 
         parsed_chunk = parse_flir_chunk(stream, chunks_count)
@@ -122,13 +143,15 @@ def parse_flir_chunk(
     #
     # Meaning: Exif APP1, 65534 long, FLIR chunk 1 out of 12
     marker = stream.read(chunk_app1_bytes_count)
-    length = int.from_bytes(stream.read(chunk_length_bytes_count), "big")
+
+    length_bytes = stream.read(chunk_length_bytes_count)
+    length = int.from_bytes(length_bytes, "big")
     length -= chunk_metadata_length
     magic_flir = stream.read(chunk_magic_bytes_count)
 
     if not (marker == app1_marker and magic_flir == magic_flir_def):
         # Seek back to just after byte b and continue searching for chunks
-        stream.seek(-chunk_partial_metadata_length, 1)
+        stream.seek(-len(marker) - len(length_bytes) - len(magic_flir), 1)
         return None
 
     stream.seek(1, 1)  # skip 1 byte, unsure what it is for
@@ -157,17 +180,27 @@ def parse_flir_chunk(
 
 def parse_thermal(
     stream: BinaryIO, records: Dict[int, Tuple[int, int, int, int]]
-) -> Array[float, ..., ...]:
+) -> FlyrThermogram:
     # FIXME Support formats other than subtype == PNG
     RECORD_IDX_RAW_DATA = 1
+    RECORD_IDX_EMBEDDED_IMAGE = 14
     RECORD_IDX_CAMERA_INFO = 32
 
     raw_data_md = records[RECORD_IDX_RAW_DATA]
-    camera_info_md = records[RECORD_IDX_CAMERA_INFO]
     width, height, raw_data = parse_raw_data(stream, raw_data_md)
+
+    embedded_image = None
+    try:
+        embedded_image_md = records[RECORD_IDX_EMBEDDED_IMAGE]
+        _, _, embedded_image = parse_embedded_image(stream, embedded_image_md)
+    except:
+        pass
+
+    camera_info_md = records[RECORD_IDX_CAMERA_INFO]
     camera_info = parse_camera_info(stream, camera_info_md)
 
-    return raw2temp(raw_data, **camera_info)
+    thermogram = FlyrThermogram(raw_data, camera_info, optical=embedded_image)
+    return thermogram
 
 
 def parse_flir_app1(stream: BinaryIO) -> Dict[int, Tuple[int, int, int, int]]:
@@ -248,13 +281,9 @@ def parse_flir_record_metadata(
     return (entry, record_type, record_offset, record_length)
 
 
-def extract_optical(stream: BinaryIO) -> Array[int, ..., ..., ...]:
-    pass  # TODO write optical data extractor
-
-
 def parse_raw_data(
     stream: BinaryIO, metadata: Tuple[int, int, int, int]
-) -> Tuple[int, int, Array[int, ..., ...]]:
+) -> Tuple[int, int, Array[np.uint8, ..., ...]]:
     (entry_idx, _, offset, length) = metadata
     stream.seek(offset)
 
@@ -270,7 +299,7 @@ def parse_raw_data(
     # print(width, height, type)
     # stream.seek(offset + 2*16)
 
-    # Read the bytes with the raw thermal data
+    # Read the bytes with the raw thermal data and decode using PIL
     thermal_bytes = stream.read(length)
     thermal_stream = BytesIO(thermal_bytes)
     thermal_img = Image.open(thermal_stream)
@@ -278,9 +307,9 @@ def parse_raw_data(
 
     # Check shape
     if thermal_np.shape != (height, width):
-        raise ValueError(
-            f"Invalid FLIR: metadata's width and height don't match thermal data's actual width and height ({thermal_np.shape} vs ({height}, {width})"
-        )
+        msg = "Invalid FLIR: metadata's width and height don't match thermal data's actual width and height ({} vs ({}, {})"
+        msg = msg.format(thermal_np.shape, height, width)
+        raise ValueError(msg)
 
     # FLIR PNG data is in the wrong byte order, fix that
     fix_byte_order = np.vectorize(lambda x: (x >> 8) + ((x & 0x00FF) << 8))
@@ -289,6 +318,36 @@ def parse_raw_data(
     return width, height, thermal_np
 
 
+def parse_embedded_image(
+    stream: BinaryIO, metadata: Tuple[int, int, int, int]
+) -> Tuple[int, int, Array[np.uint8, ..., ...]]:
+    (entry_idx, _, offset, length) = metadata
+    stream.seek(offset)
+
+    # from PIL import ImageFile  # Uncomment to enable loading corrupt PNGs
+    # ImageFile.LOAD_TRUNCATED_IMAGES = True
+
+    stream.seek(2, 1)  # Skip first two bytes, TODO Explain the why of the skip
+    width = int.from_bytes(stream.read(2), "little")
+    height = int.from_bytes(stream.read(2), "little")
+
+    stream.seek(offset + 2 * 16)  # TODO document why 2 * 16
+
+    # Read the bytes with the embedded image and decode using PIL
+    optical_bytes = stream.read(length)
+    optical_stream = BytesIO(optical_bytes)
+    optical_img = Image.open(optical_stream)
+    optical_np = np.array(optical_img)
+
+    # Check shape
+    if optical_np.shape[:2] != (height, width):
+        msg = "Invalid FLIR: metadata's width and height don't match optical data's actual width and height ({} vs ({}, {})"
+        msg = msg.format(optical_np.shape, height, width)
+        raise ValueError(msg)
+
+    return width, height, optical_np
+
+
 def parse_camera_info(
     stream: BinaryIO, metadata: Tuple[int, int, int, int]
 ) -> Dict[str, Union[int, float]]:
@@ -311,118 +370,41 @@ def parse_camera_info(
     planck_b = stream.read(4)
     planck_f = stream.read(4)
 
+    stream.seek(offset + 112)
+    atmospheric_trans_alpha1 = stream.read(4)
+    atmospheric_trans_alpha2 = stream.read(4)
+    atmospheric_trans_beta1 = stream.read(4)
+    atmospheric_trans_beta2 = stream.read(4)
+    atmospheric_trans_x = stream.read(4)
+
     stream.seek(offset + 776)
     planck_o = stream.read(4)
     planck_r2 = stream.read(4)
 
-    import struct
-
+    # Create result dictionary
     float_from_bytes = lambda v: struct.unpack("f", v)[0]
-
-    refl_app_temp_float = float_from_bytes(refl_app_temp)
-
-    # return {
-    #    'E': 1.0, #float_from_bytes(emissivity),
-    #    'OD': 10.0, #float_from_bytes(object_distance),
-    #    'RTemp': -0.0, #refl_app_temp_float - refl_app_temp_float,
-    #    'ATemp': 10.0, #float_from_bytes(atmos_temp) - refl_app_temp_float,
-    #    'IRWTemp': 20.0, #float_from_bytes(ir_window_temp) - refl_app_temp_float,
-    #    'IRT': 1.0, #float_from_bytes(ir_window_transm),
-    #    'RH': 94, #float_from_bytes(rel_humidity),
-    #    'PR1': 15524.807, #float_from_bytes(planck_r1),
-    #    'PB': 1408.7, #float_from_bytes(planck_b),
-    #    'PF': 1, #float_from_bytes(planck_f),
-    #    'PO': -4272, #struct.unpack('i', planck_o)[0],
-    #    'PR2': 0.014428768, #float_from_bytes(planck_r2),
-    # }
-
-    return {
-        "E": float_from_bytes(emissivity),
-        "OD": float_from_bytes(object_distance),
-        "RTemp": refl_app_temp_float - refl_app_temp_float,
-        "ATemp": float_from_bytes(atmos_temp) - refl_app_temp_float,
-        "IRWTemp": float_from_bytes(ir_window_temp) - refl_app_temp_float,
-        "IRT": float_from_bytes(ir_window_transm),
-        "RH": float_from_bytes(rel_humidity) * 100,
-        "PR1": float_from_bytes(planck_r1),
-        "PB": float_from_bytes(planck_b),
-        "PF": float_from_bytes(planck_f),
-        "PO": struct.unpack("i", planck_o)[0],
-        "PR2": float_from_bytes(planck_r2),
+    camera_info = {
+        "emissivity": float_from_bytes(emissivity),
+        "object_distance": float_from_bytes(object_distance),
+        "atmospheric_temperature": float_from_bytes(atmos_temp),
+        "ir_window_temperature": float_from_bytes(ir_window_temp),
+        "ir_window_transmission": float_from_bytes(ir_window_transm),
+        "reflected_apparent_temperature": float_from_bytes(refl_app_temp),
+        "relative_humidity": float_from_bytes(rel_humidity),
+        "planck_r1": float_from_bytes(planck_r1),
+        "planck_r2": float_from_bytes(planck_r2),
+        "planck_b": float_from_bytes(planck_b),
+        "planck_f": float_from_bytes(planck_f),
+        "planck_o": struct.unpack("i", planck_o)[0],
+        "atmospheric_trans_alpha1": float_from_bytes(atmospheric_trans_alpha1),
+        "atmospheric_trans_alpha2": float_from_bytes(atmospheric_trans_alpha2),
+        "atmospheric_trans_beta1": float_from_bytes(atmospheric_trans_beta1),
+        "atmospheric_trans_beta2": float_from_bytes(atmospheric_trans_beta2),
+        "atmospheric_trans_x": float_from_bytes(atmospheric_trans_x),
     }
 
-
-def raw2temp(
-    raw,
-    E=1,
-    OD=1,
-    RTemp=20,
-    ATemp=20,
-    IRWTemp=20,
-    IRT=1,
-    RH=50,
-    PR1=21106.77,
-    PB=1501,
-    PF=1,
-    PO=-7340,
-    PR2=0.012545258,
-):
-    """
-    convert raw values from the flir sensor to temperatures in C
-    # this calculation has been ported to python from
-    # https://github.com/gtatters/Thermimage/blob/master/R/raw2temp.R
-    # a detailed explanation of what is going on here can be found there
-    """
-
-    # constants
-    ATA1 = 0.006569
-    ATA2 = 0.01262
-    ATB1 = -0.002276
-    ATB2 = -0.00667
-    ATX = 1.9
-
-    # transmission through window (calibrated)
-    emiss_wind = 1 - IRT
-    refl_wind = 0
-
-    # transmission through the air
-    h2o = (RH / 100) * exp(
-        1.5587
-        + 0.06939 * (ATemp)
-        - 0.00027816 * (ATemp) ** 2
-        + 0.00000068455 * (ATemp) ** 3
-    )
-    tau1 = ATX * exp(-sqrt(OD / 2) * (ATA1 + ATB1 * sqrt(h2o))) + (1 - ATX) * exp(
-        -sqrt(OD / 2) * (ATA2 + ATB2 * sqrt(h2o))
-    )
-    tau2 = ATX * exp(-sqrt(OD / 2) * (ATA1 + ATB1 * sqrt(h2o))) + (1 - ATX) * exp(
-        -sqrt(OD / 2) * (ATA2 + ATB2 * sqrt(h2o))
-    )
-
-    # radiance from the environment
-    raw_refl1 = PR1 / (PR2 * (exp(PB / (RTemp + 273.15)) - PF)) - PO
-    raw_refl1_attn = (1 - E) / E * raw_refl1
-    raw_atm1 = PR1 / (PR2 * (exp(PB / (ATemp + 273.15)) - PF)) - PO
-    raw_atm1_attn = (1 - tau1) / E / tau1 * raw_atm1
-    raw_wind = PR1 / (PR2 * (exp(PB / (IRWTemp + 273.15)) - PF)) - PO
-    raw_wind_attn = emiss_wind / E / tau1 / IRT * raw_wind
-    raw_refl2 = PR1 / (PR2 * (exp(PB / (RTemp + 273.15)) - PF)) - PO
-    raw_refl2_attn = refl_wind / E / tau1 / IRT * raw_refl2
-    raw_atm2 = PR1 / (PR2 * (exp(PB / (ATemp + 273.15)) - PF)) - PO
-    raw_atm2_attn = (1 - tau2) / E / tau1 / IRT / tau2 * raw_atm2
-
-    raw_obj = (
-        raw / E / tau1 / IRT / tau2
-        - raw_atm1_attn
-        - raw_atm2_attn
-        - raw_wind_attn
-        - raw_refl1_attn
-        - raw_refl2_attn
-    )
-
-    # temperature from radiance
-    temp_celcius = PB / np.log(PR1 / (PR2 * (raw_obj + PO)) + PF) - 273.15
-    return temp_celcius
+    # Filter out zero values
+    return {k: v for k, v in camera_info.items() if v}
 
 
 if __name__ == "__main__":
@@ -434,5 +416,5 @@ if __name__ == "__main__":
 
     # TODO Do something useful when called as script
     file_path = sys.argv[1]
-    thermal = unpack(file_path)
-    print(thermal)
+    celsius = unpack(file_path).celsius
+    print(celsius)
